__all__ = ("EMPTY_BYTES", "EMPTY_STRING", "DEFAULT_START", "DEFAULT_STEP")

EMPTY_BYTES = bytes()
"""An empty [`bytes`][bytes] instance."""
EMPTY_STRING = str()
"""An empty [`str`][str] instance."""

DEFAULT_START = 0
"""The default start value for counting iterators."""
DEFAULT_STEP = 1
"""The default step value for counting iterators."""
